"""sonusai torchl_predict

usage: torchl_predict [-hvrw] [-i MIXID] [-m MODEL] (-k CKPT) [-b BATCH] [-t TSTEPS] INPUT ...

options:
    -h, --help
    -v, --verbose                   Be verbose.
    -i MIXID, --mixid MIXID         Mixture ID(s) to use if input is a mixture database. [default: *].
    -m MODEL, --model MODEL         Pytorch lightning model .py file path.
    -k CKPT, --checkpoint CKPT      Pytorch lightning checkpoint file with weights.
    -b BATCH, --batch BATCH         Batch size (deprecated and forced to 1). [default: 1]
    -t TSTEPS, --tsteps TSTEPS      Timesteps. If 0, dim is not included/expected in model. [default: 0]
    -r, --reset                     Reset model between each file.
    -w, --wavdbg                    Write debug .wav files of feature input, truth, and predict. [default: False]

Run pytorch lightning prediction on a SonusAI mixture database using a model and checkpoint inputs.
The torch lightning model is imported from MODEL .py file and weights loaded from checkpoint file CKPT.

Inputs:
    MODEL       Path to a .py with MyHyperModel Pytorch Lightning model class definition
    CKPT        A pytorch lightning checkpoint file with weights.
    INPUT       The input data must be one of the following:
                * directory
                  Use SonusAI mixture database directory, generate feature and truth data if not found.
                  Run prediction on the feature. The MIXID is required (or default which is *)

                * Single WAV file or glob of WAV files
                  Using the given model, generate feature data and run prediction. A model file must be
                  provided. The MIXID is ignored.

Outputs the following to tpredict-<TIMESTAMP> directory:
    <id>.h5
        dataset:    predict
    torch_predict.log

"""
from typing import Any

import torch

from sonusai import logger
from sonusai.mixture import Feature


def power_compress(x):
    real = x[..., 0]
    imag = x[..., 1]
    spec = torch.complex(real, imag)
    mag = torch.abs(spec)
    phase = torch.angle(spec)
    mag = mag ** 0.3
    real_compress = mag * torch.cos(phase)
    imag_compress = mag * torch.sin(phase)
    return torch.stack([real_compress, imag_compress], 1)


def power_uncompress(real, imag):
    spec = torch.complex(real, imag)
    mag = torch.abs(spec)
    phase = torch.angle(spec)
    mag = mag ** (1. / 0.3)
    real_compress = mag * torch.cos(phase)
    imag_compress = mag * torch.sin(phase)
    return torch.stack([real_compress, imag_compress], -1)


def main() -> None:
    from docopt import docopt

    import sonusai
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

    verbose = args['--verbose']
    mixids = args['--mixid']
    modelpath = args['--model']
    ckpt_name = args['--checkpoint']
    batch_size = args['--batch']
    timesteps = args['--tsteps']
    reset = args['--reset']
    wavdbg = args['--wavdbg']  # write .wav if true
    input_name = args['INPUT']

    from os import makedirs
    from os.path import basename
    from os.path import isdir
    from os.path import isfile
    from os.path import join
    from os.path import splitext
    from os.path import normpath
    import h5py
    from sonusai.utils import write_wav
    # from sonusai.utils import float_to_int16

    from torchinfo import summary
    from sonusai import create_file_handler
    from sonusai import initial_log_messages
    from sonusai import update_console_handler
    from sonusai.mixture import MixtureDatabase
    from sonusai.mixture import get_feature_from_audio
    from sonusai.utils import import_keras_model
    from sonusai.mixture import read_audio
    from sonusai.utils import create_ts_name
    from sonusai.data_generator import TorchFromMixtureDatabase

    if batch_size is not None:
        batch_size = int(batch_size)
    if batch_size != 1:
        batch_size = 1
        logger.info(f'For now prediction only supports batch_size = 1, forcing it to 1 now')

    if timesteps is not None:
        timesteps = int(timesteps)

    if len(input_name) == 1 and isdir(input_name[0]):
        in_basename = basename(normpath(input_name[0]))
    else:
        in_basename = ''

    output_dir = create_ts_name('tpredict-' + in_basename)
    makedirs(output_dir, exist_ok=True)

    # Setup logging file
    logger.info(f'Created output subdirectory {output_dir}')
    create_file_handler(join(output_dir, 'torchl_predict.log'))
    update_console_handler(verbose)
    initial_log_messages('torch_predict')
    logger.info(f'torch    {torch.__version__}')

    try:
        checkpoint = torch.load(ckpt_name, map_location=lambda storage, loc: storage)
        # Note other possible way to load model (not working)
        # model = litemodule.L.LightningModule.load_from_checkpoint(ckpt_name, **hparams)
    except Exception as e:
        logger.exception(f'Error: could not load checkpoint from {ckpt_name}: {e}')
        raise SystemExit(1)

    # Import model definition file
    model_base = basename(modelpath)
    model_root = splitext(model_base)[0]
    logger.info(f'Importing {modelpath}')
    litemodule = import_keras_model(modelpath)

    if 'hyper_parameters' in checkpoint:
        hparams = checkpoint['hyper_parameters']
        logger.info(f'Found checkpoint file with hyper-params named {checkpoint["hparams_name"]} '
                    f'with {len(hparams)} total hparams.')
        if batch_size is not None and hparams['batch_size'] != batch_size:
            if batch_size != 1:
                batch_size = 1
                logger.info(f'For now prediction only supports batch_size = 1, forcing it to 1 now')
            logger.info(f'Overriding batch_size: default = {hparams["batch_size"]}; specified = {batch_size}.')
            hparams["batch_size"] = batch_size

        if timesteps is not None:
            if hparams['timesteps'] == 0 and timesteps != 0:
                logger.warning(f'Model does not contain timesteps; ignoring override.')
                timesteps = 0

            if hparams['timesteps'] != 0 and timesteps == 0:
                logger.warning(f'Model contains timesteps; ignoring override, using model default.')
                timesteps = hparams['timesteps']

            if hparams['timesteps'] != timesteps:
                logger.info(f'Overriding timesteps: default = {hparams["timesteps"]}; specified = {timesteps}.')
                hparams['timesteps'] = timesteps

        logger.info(f'Building model with hparams and batch_size={batch_size}, timesteps={timesteps}')
        try:
            model = litemodule.MyHyperModel(**hparams)  # use hparams
            # litemodule.MyHyperModel.load_from_checkpoint(ckpt_name, **hparams)
        except Exception as e:
            logger.exception(f'Error: model build (MyHyperModel) in {model_base} failed: {e}')
            raise SystemExit(1)
    else:
        logger.info(f'Warning: found checkpoint with no hyper-parameters, building model with defaults')
        try:
            tmp = litemodule.MyHyperModel()  # use default hparams
        except Exception as e:
            logger.exception(f'Error: model build (MyHyperModel) in {model_base} failed: {e}')
            raise SystemExit(1)

        if batch_size is not None:
            if tmp.batch_size != batch_size:
                logger.info(f'Overriding batch_size: default = {tmp.batch_size}; specified = {batch_size}.')
        else:
            batch_size = tmp.batch_size  # inherit

        if timesteps is not None:
            if tmp.timesteps == 0 and timesteps != 0:
                logger.warning(f'Model does not contain timesteps; ignoring override.')
                timesteps = 0

            if tmp.timesteps != 0 and timesteps == 0:
                logger.warning(f'Model contains timesteps; ignoring override.')
                timesteps = tmp.timesteps

            if tmp.timesteps != timesteps:
                logger.info(f'Overriding timesteps: default = {tmp.timesteps}; specified = {timesteps}.')
        else:
            timesteps = tmp.timesteps

        logger.info(f'Building model with default hparams and batch_size= {batch_size}, timesteps={timesteps}')
        model = litemodule.MyHyperModel(timesteps=timesteps, batch_size=batch_size)

    logger.info('')
    logger.info(summary(model))
    logger.info('')
    logger.info(f'feature       {model.feature}')
    logger.info(f'num_classes   {model.num_classes}')
    logger.info(f'batch_size    {model.batch_size}')
    logger.info(f'timesteps     {model.timesteps}')
    logger.info(f'flatten       {model.flatten}')
    logger.info(f'add1ch        {model.add1ch}')
    logger.info(f'truth_mutex   {model.truth_mutex}')
    logger.info(f'input_shape   {model.input_shape}')
    logger.info('')
    model.load_state_dict(checkpoint["state_dict"])
    model.eval()

    logger.info('')
    # Load mixture database and setup dataloader
    if len(input_name) == 1 and isdir(input_name[0]):  # Single path to mixdb subdir
        input_name = input_name[0]
        logger.info(f'Loading mixture database from {input_name}')
        mixdb = MixtureDatabase(config=input_name)
        logger.info(f'Sonusai mixture db: found {len(mixdb.mixtures)} mixtures with {mixdb.num_classes} classes')

        if mixdb.feature != model.feature:
            logger.exception(f'Feature in mixture database {mixdb.feature} does not match feature in model')
            raise SystemExit(1)

        # TBD check num_classes ??

        p_mixids = mixdb.mixids_to_list(mixids)
        sampler = None
        p_datagen = TorchFromMixtureDatabase(mixdb=mixdb,
                                             mixids=p_mixids,
                                             batch_size=model.batch_size,
                                             cut_len=0,
                                             flatten=model.flatten,
                                             add1ch=model.add1ch,
                                             random_cut=False,
                                             sampler=sampler,
                                             drop_last=False,
                                             num_workers=0)

        if wavdbg:  # setup for wav write if enabled
            # Info needed to setup inverse transform
            from pyaaware import FeatureGenerator
            from pyaaware import TorchInverseTransform
            from torchaudio import save
            # from sonusai.utils import write_wav

            half = model.num_classes // 2
            fg = FeatureGenerator(feature_mode=model.feature,
                                  num_classes=model.num_classes,
                                  truth_mutex=model.truth_mutex)
            itf = TorchInverseTransform(N=fg.itransform_N,
                                        R=fg.itransform_R,
                                        bin_start=fg.bin_start,
                                        bin_end=fg.bin_end,
                                        ttype=fg.itransform_ttype)

            if mixdb.targets[0].truth_settings[0].function == 'target_f' or \
                    mixdb.targets[0].truth_settings[0].function == 'target_mixture_f':
                enable_truth_wav = True
            else:
                enable_truth_wav = False

            if mixdb.targets[0].truth_settings[0].function == 'target_mixture_f':
                enable_mix_wav = True
            else:
                enable_mix_wav = False

        if reset:
            logger.info(f'Running {len(mixdb.mixtures)} mixtures individually with model reset ...')
            for idx, val in enumerate(p_datagen):
                # truth = val[0]
                feature = val[1]
                with torch.no_grad():
                    ypred = model(feature)
                output_name = join(output_dir, mixdb.mixtures[idx].name)
                pdat = ypred.detach().numpy()
                if timesteps > 0:
                    logger.debug(f'In and out tsteps: {feature.shape[1]},{pdat.shape[1]}')
                logger.debug(f'Writing predict shape {pdat.shape} to {output_name}')
                with h5py.File(output_name, 'a') as f:
                    if 'predict' in f:
                        del f['predict']
                    f.create_dataset('predict', data=pdat)

                if wavdbg:
                    owav_base = splitext(output_name)[0]
                    tmp = torch.complex(ypred[..., :half], ypred[..., half:]).permute(2, 0, 1).detach()
                    predwav, _ = itf.execute_all(tmp)
                    # predwav, _ = calculate_audio_from_transform(tmp, itf, trim=True)
                    save(owav_base + '.wav', predwav.permute([1, 0]), 16000, encoding='PCM_S', bits_per_sample=16)
                    if enable_truth_wav:
                        # Note this support truth type target_f and target_mixture_f
                        tmp = torch.complex(val[0][..., :half], val[0][..., half:2 * half]).permute(2, 0, 1).detach()
                        truthwav, _ = itf.execute_all(tmp)
                        save(owav_base + '_truth.wav', truthwav.permute([1, 0]), 16000, encoding='PCM_S',
                             bits_per_sample=16)

                    if enable_mix_wav:
                        tmp = torch.complex(val[0][..., 2 * half:3 * half], val[0][..., 3 * half:]).permute(2, 0, 1)
                        mixwav, _ = itf.execute_all(tmp.detach())
                        save(owav_base + "_mix.wav", mixwav.permute([1, 0]), 16000, encoding='PCM_S',
                             bits_per_sample=16)
                        # write_wav(owav_base + "_truth.wav", truthwav, 16000)

        else:
            logger.info(f'Running {len(mixdb.mixtures)} mixtures with model builtin prediction loop ...')
            from lightning.pytorch import Trainer
            trainer = Trainer(default_root_dir=output_dir,
                              accelerator='auto')  # prints avail GPU, TPU, IPU, HPU and selected device
            # logger.info(f'Strategy: {trainer.strategy.strategy_name}')  # doesn't work for ddp strategy
            logger.info(f'Accelerator stats: {trainer.accelerator.get_device_stats(device=None)}')
            logger.info(f'World size: {trainer.world_size}')
            logger.info(f'Nodes: {trainer.num_nodes}')
            logger.info(f'Devices: {trainer.accelerator.auto_device_count()}')

            # Use builtin lightning prediction loop, returns a list
            # predictions = trainer.predict(model, p_datagen)  # standard method, but no support distributed
            trainer.predict(model, p_datagen)
            predictions = model.predict_outputs
            # all_predictions = torch.cat(predictions)   #  predictions = torch.cat(predictions).cpu()
            if trainer.world_size > 1:
                # print(f'Predictions returned: {len(all_predictions)}')
                ddp_max_mem = torch.cuda.max_memory_allocated(trainer.local_rank) / 1000
                logger.info(f"GPU {trainer.local_rank} max memory using DDP: {ddp_max_mem:.2f} MB")
                gathered = [None] * torch.distributed.get_world_size()
                torch.distributed.all_gather_object(gathered, predictions)
                torch.distributed.barrier()
                if not trainer.is_global_zero:
                    return
                predictions = sum(gathered, [])
                if trainer.global_rank == 0:
                    logger.info(f"All predictions gathered: {len(predictions)}")

            logger.info(f'Predictions returned: {len(predictions)}')
            for idx, mixid in enumerate(p_mixids):
                # print(f'{idx}, {mixid}')
                output_name = join(output_dir, mixdb.mixtures[mixid].name)
                pdat = predictions[idx].cpu().numpy()
                logger.debug(f'Writing predict shape {pdat.shape} to {output_name}')
                with h5py.File(output_name, 'a') as f:
                    if 'predict' in f:
                        del f['predict']
                    f.create_dataset('predict', data=pdat)

                if wavdbg:
                    owav_base = splitext(output_name)[0]
                    tmp = torch.complex(predictions[idx][..., :half], predictions[idx][..., half:]).permute(2, 1, 0)
                    predwav, _ = itf.execute_all(tmp.squeeze().detach().numpy())
                    write_wav(owav_base + ".wav", predwav.detach().numpy(), 16000)

        logger.info(f'Saved results to {output_dir}')
        return

        # if reset:
        #     # reset mode cycles through each file one at a time
        #     for mixid in mixids:
        #         feature, _ = mixdb.mixture_ft(mixid)
        #         if feature.shape[0] > 2500:
        #             print(f'Trimming input frames from {feature.shape[0]} to {2500},')
        #             feature = feature[0:2500,::]
        #         half = feature.shape[-1] // 2
        #         noisy_spec_cmplx = torch.complex(torch.tensor(feature[..., :half]),
        #                                          torch.tensor(feature[..., half:])).to(device)
        #         del feature
        #
        #         predict = _pad_and_predict(built_model=model, feature=noisy_spec_cmplx)
        #         del noisy_spec_cmplx
        #
        #         audio_est = torch_istft_olsa_hanns(predict, mixdb.it_config.N, mixdb.it_config.R).cpu()
        #         del predict
        #         output_name = join(output_dir, splitext(mixdb.mixtures[mixid].name)[0]+'.wav')
        #         print(f'Saving prediction to {output_name}')
        #         write_wav(name=output_name, audio=float_to_int16(audio_est.detach().numpy()).transpose())
        #
        #         torch.cuda.empty_cache()
        #
        #         # TBD .h5 predict file optional output file
        #         # output_name = join(output_dir, mixdb.mixtures[mixid].name)
        #         # with h5py.File(output_name, 'a') as f:
        #         #     if 'predict' in f:
        #         #         del f['predict']
        #         #     f.create_dataset(name='predict', data=predict)
        #
        # else:
        #     # Run all data at once using a data generator
        #     feature = KerasFromH5(mixdb=mixdb,
        #                           mixids=mixids,
        #                           batch_size=hypermodel.batch_size,
        #                           timesteps=hypermodel.timesteps,
        #                           flatten=hypermodel.flatten,
        #                           add1ch=hypermodel.add1ch)
        #
        #     predict = built_model.predict(feature, batch_size=hypermodel.batch_size, verbose=1)
        #     predict, _ = reshape_outputs(predict=predict, timesteps=hypermodel.timesteps)
        #
        #     # Write data to separate files
        #     for idx, mixid in enumerate(mixids):
        #         output_name = join(output_dir, mixdb.mixtures[mixid].name)
        #         with h5py.File(output_name, 'a') as f:
        #             if 'predict' in f:
        #                 del f['predict']
        #             f.create_dataset('predict', data=predict[feature.file_indices[idx]])
        #
        # logger.info(f'Saved results to {output_dir}')
        # return

    if not all(isfile(file) and splitext(file)[1] == '.wav' for file in input_name):
        logger.exception(f'Do not know how to process input from {input_name}')
        raise SystemExit(1)

    logger.info(f'Run prediction on {len(input_name):,} WAV files')
    for file in input_name:
        # Convert WAV to feature data
        audio = read_audio(file)
        feature = get_feature_from_audio(audio=audio, feature=model.feature)

        # feature, predict = _pad_and_predict(hypermodel=hypermodel,
        #                                     built_model=built_model,
        #                                     feature=feature,
        #                                     frames_per_batch=frames_per_batch)

        # clean = torch_istft_olsa_hanns(clean_spec_cmplx, mixdb.ift_config.N, mixdb.ift_config.R)

        output_name = join(output_dir, splitext(basename(file))[0] + '.h5')
        with h5py.File(output_name, 'a') as f:
            if 'feature' in f:
                del f['feature']
            f.create_dataset(name='feature', data=feature)

            # if 'predict' in f:
            #     del f['predict']
            # f.create_dataset(name='predict', data=predict)

    logger.info(f'Saved results to {output_dir}')
    del model


def _pad_and_predict(built_model: Any, feature: Feature) -> torch.Tensor:
    """
    Run prediction on feature [frames,1,bins*2] (stacked complex numpy array, stride/tsteps=1)
    Returns predict output [batch,frames,bins] in complex torch.tensor
    """
    noisy_spec = power_compress(torch.view_as_real(torch.from_numpy(feature).permute(1, 0, 2)))
    # print(f'noisy_spec type {type(noisy_spec_cmplx)}')
    # print(f'noisy_spec dtype {noisy_spec_cmplx.dtype}')
    # print(f'noisy_spec size {noisy_spec_cmplx.shape}')
    with torch.no_grad():
        est_real, est_imag = built_model(noisy_spec)  # expects in size [batch, 2, tsteps, bins]
    est_real, est_imag = est_real.permute(0, 1, 3, 2), est_imag.permute(0, 1, 3, 2)
    est_spec_uncompress = torch.view_as_complex(power_uncompress(est_real, est_imag).squeeze(1))
    # inv tf want [ch,frames,bins] complex (synonymous with [batch,tsteps,bins]), keep as torch.tensor
    predict = est_spec_uncompress.permute(0, 2, 1)  # .detach().numpy()

    return predict


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        exit()



from typing import Deque
from typing import List

from sonusai.mixture.mixdb import MixtureDatabase
from sonusai.mixture.types import Augmentations
from sonusai.mixture.types import AugmentedTarget
from sonusai.mixture.types import AugmentedTargets
from sonusai.mixture.types import TargetFile
from sonusai.mixture.types import TargetFiles


def get_augmented_targets(target_files: TargetFiles, target_augmentations: Augmentations) -> AugmentedTargets:
    from sonusai.mixture import get_augmentation_indices_for_mixup
    from sonusai.mixture import get_mixups

    mixups = get_mixups(target_augmentations)

    augmented_targets: AugmentedTargets = []
    for mixup in mixups:
        augmentation_indices = get_augmentation_indices_for_mixup(target_augmentations, mixup)
        for target_index in range(len(target_files)):
            for augmentation_index in augmentation_indices:
                augmented_targets.append(AugmentedTarget(target_file_index=target_index,
                                                         target_augmentation_index=augmentation_index))

    return augmented_targets


def get_truth_indices_for_target(target: TargetFile) -> List[int]:
    """Get a list of truth indices for a given target."""
    index = [truth_setting.index for truth_setting in target.truth_settings]

    # flatten, uniquify, and sort
    return sorted(list(set([item for sublist in index for item in sublist])))


def get_truth_indices_for_augmented_target(augmented_target: AugmentedTarget, targets: TargetFiles) -> List[int]:
    return get_truth_indices_for_target(targets[augmented_target.target_file_index])


def get_mixup_for_augmented_target(augmented_target: AugmentedTarget, augmentations: Augmentations) -> int:
    return augmentations[augmented_target.target_augmentation_index].mixup


def get_target_indices_for_truth_index(targets: TargetFiles,
                                       truth_index: int,
                                       allow_multiple: bool = False) -> List[int]:
    """Get a list of target indices containing the given truth index.

    If allow_multiple is True, then include targets that contain multiple truth indices.
    """
    target_indices = set()
    for target_index, target in enumerate(targets):
        indices = get_truth_indices_for_target(target)
        if len(indices) == 1 or allow_multiple:
            for index in indices:
                if index == truth_index + 1:
                    target_indices.add(target_index)

    return sorted(list(target_indices))


def get_augmented_target_indices_for_truth_index(augmented_targets: AugmentedTargets,
                                                 targets: TargetFiles,
                                                 augmentations: Augmentations,
                                                 truth_index: int,
                                                 mixup: int,
                                                 allow_multiple: bool = False) -> List[int]:
    """Get a list of augmented target indices containing the given truth index.

    If allow_multiple is True, then include targets that contain multiple truth indices.
    """
    augmented_target_indices = set()
    for augmented_target_index, augmented_target in enumerate(augmented_targets):
        if get_mixup_for_augmented_target(augmented_target=augmented_target, augmentations=augmentations) == mixup:
            indices = get_truth_indices_for_augmented_target(augmented_target=augmented_target, targets=targets)
            if len(indices) == 1 or allow_multiple:
                for index in indices:
                    if index == truth_index + 1:
                        augmented_target_indices.add(augmented_target_index)

    return sorted(list(augmented_target_indices))


def get_augmented_target_indices_by_class(mixdb: MixtureDatabase,
                                          augmented_targets: AugmentedTargets,
                                          mixup: int) -> List[List[int]]:
    num_classes = mixdb.num_classes
    if mixdb.truth_mutex:
        num_classes -= 1

    indices = []
    for idx in range(num_classes):
        indices.append(
            get_augmented_target_indices_for_truth_index(augmented_targets=augmented_targets,
                                                         targets=mixdb.targets,
                                                         augmentations=mixdb.target_augmentations,
                                                         truth_index=idx,
                                                         mixup=mixup))
    return indices


def get_augmented_target_indices_for_mixup(mixdb: MixtureDatabase,
                                           augmented_targets: AugmentedTargets,
                                           mixup: int) -> List[List[int]]:
    from collections import deque
    from random import shuffle

    from sonusai import SonusAIError

    mixup_indices = []

    if mixup == 1:
        for index, augmented_target in enumerate(augmented_targets):
            if get_mixup_for_augmented_target(augmented_target=augmented_target,
                                              augmentations=mixdb.target_augmentations) == 1:
                mixup_indices.append([index])
        return mixup_indices

    augmented_target_indices_by_class = get_augmented_target_indices_by_class(mixdb=mixdb,
                                                                              augmented_targets=augmented_targets,
                                                                              mixup=mixup)

    if mixup > mixdb.num_classes:
        raise SonusAIError(
            f'Specified mixup, {mixup}, is greater than the number of classes, {mixdb.num_classes}')

    de: Deque[int] = deque()

    # Keep looping until not enough targets remain for mixup
    while sum([1 for x in augmented_target_indices_by_class if x]) >= mixup:
        # Need more class indices?
        if len(de) < mixup:
            # Only choose classes that still have data
            counts = [len(item) for item in augmented_target_indices_by_class]
            # Need to subtract out indices already in the deque
            for idx in de:
                counts[idx] -= 1
            indices = [idx for idx, val in enumerate(counts) if val > 0]
            shuffle(indices)
            # Keep shuffling if the deque is not empty and the first new index matches the last item
            # (so that a class does not appear twice in a mixup)
            while de and indices[0] == de[-1]:
                shuffle(indices)
            for index in indices:
                de.append(index)

        class_indices = [de.popleft() for _ in range(mixup)]

        target_indices = []
        for class_index in class_indices:
            target_indices.append(augmented_target_indices_by_class[class_index].pop())

        mixup_indices.append(target_indices)

    return mixup_indices



from typing import Tuple

import sox
from pyaaware import ForwardTransform
from pyaaware import InverseTransform

from sonusai.mixture.types import AudioF
from sonusai.mixture.types import AudioT
from sonusai.mixture.types import EnergyT
from sonusai.mixture.types import ImpulseResponseRawData
from sonusai.mixture.types import Location


def get_next_noise(audio: AudioT, offset: int, length: int) -> AudioT:
    """Get next sequence of noise data from noise audio file

    :param audio: Overall noise audio (entire file's worth of data)
    :param offset: Starting sample
    :param length: Number of samples to get
    :return: Sequence of noise audio data
    """
    import numpy as np

    return np.take(audio, range(offset, offset + length), mode='wrap')


def read_ir(audio_file: Location) -> ImpulseResponseRawData:
    """Read impulse response data

    :param audio_file: Audio file name
    :return: ImpulseResponseRawData object
    """
    import tempfile

    import numpy as np
    from scipy.io import wavfile

    from sonusai import SonusAIError
    from sonusai.mixture import ImpulseResponseData
    from sonusai.mixture import ImpulseResponseRawData
    from sonusai.mixture import apply_ir
    from sonusai.mixture import tokenized_expand

    expanded_name, _ = tokenized_expand(audio_file)

    # Read impulse response data from audio file
    try:
        sample_rate, data = wavfile.read(expanded_name)
    except Exception as e:
        if audio_file != expanded_name:
            raise SonusAIError(f'Error reading {audio_file} (expanded: {expanded_name}): {e}')
        else:
            raise SonusAIError(f'Error reading {audio_file}: {e}')

    # Normalize to -20 dBFS to avoid clipping when applying IR
    max_data = max(abs(data)) * 10
    data = (data / max_data).astype(np.float32)

    # Find offset to align convolved audio with original
    temp = tempfile.NamedTemporaryFile(mode='w+t')
    try:
        for d in data:
            temp.write(f'{d:f}\n')
        temp.seek(0)

        ir = ImpulseResponseData(
            name=audio_file,
            sample_rate=sample_rate,
            offset=0,
            length=len(data),
            coefficients_file=temp.name)
        x = np.zeros((ir.length,), dtype=np.float32)
        x[0] = 1
        y = list(apply_ir(x, ir))
    finally:
        temp.close()

    return ImpulseResponseRawData(
        name=audio_file,
        sample_rate=sample_rate,
        offset=y.index(max(y)),
        filter=data)


def save_ir_data(name: Location, data: AudioT) -> None:
    with open(name, 'w+t') as f:
        for d in data:
            f.write(f'{d:f}\n')


def read_audio(name: Location) -> AudioT:
    """Read audio data from a file

    :param name: File name
    :return: Array of time domain audio data
    """
    from typing import Any

    import numpy as np

    from sonusai import SonusAIError
    from sonusai.mixture import BIT_DEPTH
    from sonusai.mixture import CHANNEL_COUNT
    from sonusai.mixture import ENCODING
    from sonusai.mixture import SAMPLE_RATE
    from sonusai.mixture import tokenized_expand

    def encode_output(buffer: Any) -> np.ndarray:
        if BIT_DEPTH == 8:
            return np.frombuffer(buffer, dtype=np.int8)

        if BIT_DEPTH == 16:
            return np.frombuffer(buffer, dtype=np.int16)

        if BIT_DEPTH == 24:
            return np.frombuffer(buffer, dtype=np.int32)

        if BIT_DEPTH == 32:
            if ENCODING == 'floating-point':
                return np.frombuffer(buffer, dtype=np.float32)
            return np.frombuffer(buffer, dtype=np.int32)

        if BIT_DEPTH == 64:
            return np.frombuffer(buffer, dtype=np.float64)

        raise SonusAIError(f'Invalid BIT_DEPTH {BIT_DEPTH}')

    expanded_name, _ = tokenized_expand(name)

    try:
        # Read in and convert to desired format
        # NOTE: pysox format transformations do not handle encoding properly; need to use direct call to sox instead
        args = [
            '-D',
            '-G',
            expanded_name,
            '-t', 'raw',
            '-r', str(SAMPLE_RATE),
            '-b', str(BIT_DEPTH),
            '-c', str(CHANNEL_COUNT),
            '-e', ENCODING,
            '-',
            'remix', '1',
        ]
        status, out, err = sox.core.sox(args, None, False)
        if status != 0:
            raise SonusAIError(f'sox stdout: {out}\nsox stderr: {err}')

        return encode_output(out)

    except Exception as e:
        if name != expanded_name:
            raise SonusAIError(f'Error reading {name} (expanded: {expanded_name}):\n{e}')
        else:
            raise SonusAIError(f'Error reading {name}:\n{e}')


def calculate_transform_from_audio(audio: AudioT,
                                   transform: ForwardTransform) -> Tuple[AudioF, EnergyT]:
    """Apply forward transform to input audio data to generate transform data

    :param audio: Time domain data [samples]
    :param transform: ForwardTransform object
    :return: Frequency domain data [frames, bins], Energy [frames]
    """
    f, e = transform.execute_all(audio)
    return f.transpose(), e


def calculate_audio_from_transform(data: AudioF,
                                   transform: InverseTransform,
                                   trim: bool = True) -> Tuple[AudioT, EnergyT]:
    """Apply inverse transform to input transform data to generate audio data

    :param data: Frequency domain data [frames, bins]
    :param transform: InverseTransform object
    :param trim: Removes starting samples so output waveform will be time-aligned with input waveform to the transform
    :return: Time domain data [samples], Energy [frames]
    """
    t, e = transform.execute_all(data.transpose())
    if trim:
        t = t[transform.N - transform.R:]

    return t, e


def get_duration(audio: AudioT) -> float:
    """Get duration of audio in seconds

    :param audio: Time domain data [samples]
    :return: Duration of audio in seconds
    """
    from sonusai.mixture import SAMPLE_RATE

    return len(audio) / SAMPLE_RATE


class Transformer(sox.Transformer):
    """Override certain sox.Transformer methods
    """

    def fir(self, coefficients):
        """Use SoX’s FFT convolution engine with given FIR filter coefficients.

        The SonusAI override allows coefficients to be either a list of numbers
        or a string containing a text file with the coefficients.

        Parameters
        ----------
        coefficients : list or str
            fir filter coefficients

        """
        from sonusai import SonusAIError

        if not isinstance(coefficients, list) and not isinstance(coefficients, str):
            raise SonusAIError("coefficients must be a list or a str.")

        if isinstance(coefficients, list) and not all([sox.core.is_number(c) for c in coefficients]):
            raise SonusAIError("coefficients list must be numbers.")

        effect_args = ['fir']
        if isinstance(coefficients, list):
            effect_args.extend(['{:f}'.format(c) for c in coefficients])
        else:
            effect_args.append(coefficients)

        self.effects.extend(effect_args)
        self.effects_log.append('fir')

        return self

    def tempo(self, factor, audio_type=None, quick=False):
        """Time stretch audio without changing pitch.

        This effect uses the WSOLA algorithm. The audio is chopped up into
        segments which are then shifted in the time domain and overlapped
        (cross-faded) at points where their waveforms are most similar as
        determined by measurement of least squares.

        The SonusAI override does not generate a warning for small factors.
        The sox.Transformer's implementation of stretch does not invert
        the factor even though it says that it does; this invalidates the
        factor size check and produces the wrong result.

        Parameters
        ----------
        factor : float
            The ratio of new tempo to the old tempo.
            For ex. 1.1 speeds up the tempo by 10%; 0.9 slows it down by 10%.
        audio_type : str
            Type of audio, which optimizes algorithm parameters. One of:
             * m : Music,
             * s : Speech,
             * l : Linear (useful when factor is close to 1),
        quick : bool, default=False
            If True, this effect will run faster but with lower sound quality.

        See Also
        --------
        stretch, speed, pitch

        """
        from sox.core import is_number

        from sonusai import SonusAIError
        from sonusai import logger

        if not is_number(factor) or factor <= 0:
            raise SonusAIError('factor must be a positive number')

        if factor < 0.5 or factor > 2:
            logger.warning('Using an extreme time stretching factor. Quality of results will be poor')

        if audio_type not in [None, 'm', 's', 'l']:
            raise SonusAIError("audio_type must be one of None, 'm', 's', or 'l'.")

        if not isinstance(quick, bool):
            raise SonusAIError('quick must be a boolean')

        effect_args = ['tempo']

        if quick:
            effect_args.append('-q')

        if audio_type is not None:
            effect_args.append('-{}'.format(audio_type))

        effect_args.append('{:f}'.format(factor))

        self.effects.extend(effect_args)
        self.effects_log.append('tempo')

        return self


def validate_input_file(input_filepath: str) -> None:
    from os.path import exists
    from os.path import splitext

    from sox.core import VALID_FORMATS

    from sonusai import SonusAIError

    if not exists(input_filepath):
        raise SonusAIError(f'input_filepath {input_filepath} does not exist.')

    ext = splitext(input_filepath)[1][1:].lower()
    if ext not in VALID_FORMATS:
        raise SonusAIError(f'This installation of SoX cannot process .{ext} files')



from typing import Optional
from typing import Tuple
from typing import Type

import numpy as np
from torch.utils.data import DataLoader
from torch.utils.data import Dataset
from torch.utils.data import Sampler

from sonusai.mixture import GeneralizedIDs
from sonusai.mixture import MixtureDatabase


class MixtureDatabaseDataset(Dataset):
    """Generates a PyTorch dataset from a SonusAI mixture database
    """

    def __init__(self,
                 mixdb: MixtureDatabase,
                 mixids: GeneralizedIDs,
                 cut_len: int,
                 flatten: bool,
                 add1ch: bool,
                 random_cut: bool = True):
        """Initialization
        """
        self.mixdb = mixdb
        self.mixids = self.mixdb.mixids_to_list(mixids)
        self.cut_len = cut_len
        self.flatten = flatten
        self.add1ch = add1ch
        self.random_cut = random_cut

    def __len__(self):
        return len(self.mixids)

    def __getitem__(self, idx: int) -> Tuple[np.ndarray, np.ndarray]:
        """Get data from one mixture
        """
        import random

        from sonusai.utils import reshape_inputs

        feature, truth = self.mixdb.mixture_ft(self.mixids[idx])
        feature, truth = reshape_inputs(feature=feature,
                                        truth=truth,
                                        batch_size=1,
                                        timesteps=0,
                                        flatten=self.flatten,
                                        add1ch=self.add1ch)

        length = feature.shape[0]

        if self.cut_len > 0:
            if length < self.cut_len:
                feature_final = []
                truth_final = []
                for _ in range(self.cut_len // length):
                    feature_final.append(feature)
                    truth_final.append(truth)
                feature_final.append(feature[: self.cut_len % length])
                truth_final.append(truth[: self.cut_len % length])
                feature = np.vstack([feature_final[i] for i in range(len(feature_final))])
                truth = np.vstack([truth_final[i] for i in range(len(truth_final))])
            else:
                if self.random_cut:
                    start = random.randint(0, length - self.cut_len)
                else:
                    start = 0
                feature = feature[start:start + self.cut_len]
                truth = truth[start:start + self.cut_len]

        return feature, truth


class AawareDataLoader(DataLoader):
    _cut_len: Optional[int] = None

    @property
    def cut_len(self) -> int:
        return self._cut_len

    @cut_len.setter
    def cut_len(self, value: int) -> None:
        self._cut_len = value


def TorchFromMixtureDatabase(mixdb: MixtureDatabase,
                             mixids: GeneralizedIDs,
                             batch_size: int,
                             flatten: bool,
                             add1ch: bool,
                             num_workers: int = 0,
                             cut_len: int = 0,
                             drop_last: bool = False,
                             shuffle: bool = False,
                             random_cut: bool = True,
                             sampler: Optional[Type[Sampler]] = None,
                             pin_memory: bool = False) -> AawareDataLoader:
    """Generates a PyTorch dataloader from a SonusAI mixture database
    """
    dataset = MixtureDatabaseDataset(mixdb=mixdb,
                                     mixids=mixids,
                                     cut_len=cut_len,
                                     flatten=flatten,
                                     add1ch=add1ch,
                                     random_cut=random_cut)

    if sampler is not None:
        my_sampler = sampler(dataset)
    else:
        my_sampler = None

    result = AawareDataLoader(dataset=dataset,
                              batch_size=batch_size,
                              pin_memory=pin_memory,
                              shuffle=shuffle,
                              sampler=my_sampler,
                              drop_last=drop_last,
                              num_workers=num_workers)

    result.cut_len = cut_len

    return result



from typing import List
from typing import Tuple

from sonusai.mixture.mixdb import MixtureDatabase
from sonusai.mixture.types import Augmentations
from sonusai.mixture.types import AugmentedTargets
from sonusai.mixture.types import NoiseFiles


def generate_mixtures(mixdb: MixtureDatabase,
                      augmented_targets: AugmentedTargets,
                      noise_files: NoiseFiles,
                      noise_augmentations: Augmentations,
                      mixups: List[int]) -> Tuple[int, int]:
    """Generate mixtures and append to mixture database

    :param mixdb: Mixture database
    :param augmented_targets: List of augmented targets
    :param noise_files: List of noise files
    :param noise_augmentations: List of noise augmentations
    :param mixups: List of mixup values
    :return: (Number of noise files used, number of noise samples used)
    """
    from sonusai import SonusAIError

    if mixdb.noise_mix_mode == 'exhaustive':
        return _exhaustive_noise_mix(mixdb=mixdb,
                                     augmented_targets=augmented_targets,
                                     noise_files=noise_files,
                                     noise_augmentations=noise_augmentations,
                                     mixups=mixups)

    if mixdb.noise_mix_mode == 'non-exhaustive':
        return _non_exhaustive_noise_mix(mixdb=mixdb,
                                         augmented_targets=augmented_targets,
                                         mixups=mixups)

    if mixdb.noise_mix_mode == 'non-combinatorial':
        return _non_combinatorial_noise_mix(mixdb=mixdb,
                                            augmented_targets=augmented_targets,
                                            mixups=mixups)

    raise SonusAIError(f'invalid noise_mix_mode: {mixdb.noise_mix_mode}')


def _exhaustive_noise_mix(mixdb: MixtureDatabase,
                          augmented_targets: AugmentedTargets,
                          noise_files: NoiseFiles,
                          noise_augmentations: Augmentations,
                          mixups: List[int]) -> Tuple[int, int]:
    """Exhaustive noise mix mode
    Use every noise/augmentation with every target/augmentation.

    :param mixdb: Mixture database
    :param augmented_targets: List of augmented targets
    :param noise_files: List of noise files
    :param noise_augmentations: List of noise augmentations
    :param mixups: List of mixup values
    :return: (Number of noise files used, number of noise samples used)
    """
    from random import randint

    import numpy as np

    from sonusai.mixture import MRecord
    from sonusai.mixture import get_augmented_target_indices_for_mixup

    used_noise_files = len(noise_files) * len(noise_augmentations)
    used_noise_samples = 0

    augmented_target_indices_for_mixups = [get_augmented_target_indices_for_mixup(mixdb=mixdb,
                                                                                  augmented_targets=augmented_targets,
                                                                                  mixup=mixup) for mixup in mixups]
    for noise_file_index in range(len(noise_files)):
        for noise_augmentation_index in range(len(noise_augmentations)):
            noise_offset = 0
            noise_length = mixdb.augmented_noise_length(noise_file_index, noise_augmentation_index)

            for augmented_target_indices_for_mixup in augmented_target_indices_for_mixups:
                for augmented_target_indices in augmented_target_indices_for_mixup:
                    (target_file_index,
                     target_augmentation_index,
                     target_length) = _get_target_info(mixdb=mixdb,
                                                       augmented_target_indices=augmented_target_indices,
                                                       augmented_targets=augmented_targets)

                    for spectral_mask_index in range(len(mixdb.spectral_masks)):
                        for snr in mixdb.all_snrs:
                            mixdb.mixtures.append(MRecord(
                                name=None,
                                target_file_index=target_file_index,
                                target_augmentation_index=target_augmentation_index,
                                noise_file_index=noise_file_index,
                                noise_offset=noise_offset,
                                noise_augmentation_index=noise_augmentation_index,
                                samples=target_length,
                                snr=snr.value,
                                spectral_mask_index=spectral_mask_index,
                                spectral_mask_seed=randint(0, np.iinfo('i').max),
                                random_snr=snr.is_random))

                            noise_offset = int((noise_offset + target_length) % noise_length)
                            used_noise_samples += target_length

    return used_noise_files, used_noise_samples


def _non_exhaustive_noise_mix(mixdb: MixtureDatabase,
                              augmented_targets: AugmentedTargets,
                              mixups: List[int]) -> Tuple[int, int]:
    """Non-exhaustive noise mix mode
    Cycle through every target/augmentation without necessarily using all
    noise/augmentation combinations (reduced data set).

    :param mixdb: Mixture database
    :param augmented_targets: List of augmented targets
    :param mixups: List of mixup values
    :return: (Number of noise files used, number of noise samples used)
    """
    from random import randint

    import numpy as np

    from sonusai.mixture import MRecord
    from sonusai.mixture import get_augmented_target_indices_for_mixup

    used_noise_files = set()
    used_noise_samples = 0
    noise_offset = 0
    noise_file_index = 0
    noise_augmentation_index = 0

    augmented_target_indices_for_mixups = [get_augmented_target_indices_for_mixup(mixdb=mixdb,
                                                                                  augmented_targets=augmented_targets,
                                                                                  mixup=mixup) for mixup in mixups]
    for mixup in augmented_target_indices_for_mixups:
        for augmented_target_indices in mixup:
            (target_file_index,
             target_augmentation_index,
             target_length) = _get_target_info(mixdb=mixdb,
                                               augmented_target_indices=augmented_target_indices,
                                               augmented_targets=augmented_targets)

            for spectral_mask_index in range(len(mixdb.spectral_masks)):
                for snr in mixdb.all_snrs:
                    used_noise_files.add(f'{noise_file_index}_{noise_augmentation_index}')
                    (noise_file_index,
                     noise_augmentation_index,
                     noise_offset) = _get_next_noise_offset(mixdb=mixdb,
                                                            target_length=target_length,
                                                            noise_file_index=noise_file_index,
                                                            noise_augmentation_index=noise_augmentation_index,
                                                            noise_offset=noise_offset)

                    mixdb.mixtures.append(MRecord(
                        name=None,
                        target_file_index=target_file_index,
                        target_augmentation_index=target_augmentation_index,
                        noise_file_index=noise_file_index,
                        noise_augmentation_index=noise_augmentation_index,
                        noise_offset=noise_offset,
                        samples=target_length,
                        snr=snr.value,
                        spectral_mask_index=spectral_mask_index,
                        spectral_mask_seed=randint(0, np.iinfo('i').max),
                        random_snr=snr.is_random))

                    noise_offset += target_length
                    used_noise_samples += target_length

    return len(used_noise_files), used_noise_samples


def _non_combinatorial_noise_mix(mixdb: MixtureDatabase,
                                 augmented_targets: AugmentedTargets,
                                 mixups: List[int]) -> Tuple[int, int]:
    """Non-combinatorial noise mix mode
    Combine a target/augmentation with a single cut of a noise/augmentation
    non-exhaustively (each target/augmentation does not use each noise/augmentation).
    Cut has random start and loop back to beginning if end of noise/augmentation is reached.

    :param mixdb: Mixture database
    :param augmented_targets: List of augmented targets
    :param mixups: List of mixup values
    :return: (Number of noise files used, number of noise samples used)
    """
    from random import choice
    from random import randint

    import numpy as np

    from sonusai.mixture import MRecord
    from sonusai.mixture import get_augmented_target_indices_for_mixup

    used_noise_files = set()
    used_noise_samples = 0
    noise_file_index = 0
    noise_augmentation_index = 0

    augmented_target_indices_for_mixups = [get_augmented_target_indices_for_mixup(mixdb=mixdb,
                                                                                  augmented_targets=augmented_targets,
                                                                                  mixup=mixup) for mixup in mixups]
    for mixup in augmented_target_indices_for_mixups:
        for augmented_target_indices in mixup:
            (target_file_index,
             target_augmentation_index,
             target_length) = _get_target_info(mixdb=mixdb,
                                               augmented_target_indices=augmented_target_indices,
                                               augmented_targets=augmented_targets)

            for spectral_mask_index in range(len(mixdb.spectral_masks)):
                for snr in mixdb.all_snrs:
                    used_noise_files.add(f'{noise_file_index}_{noise_augmentation_index}')
                    (noise_file_index,
                     noise_augmentation_index,
                     noise_length) = _get_next_noise_indices(mixdb=mixdb,
                                                             noise_file_index=noise_file_index,
                                                             noise_augmentation_index=noise_augmentation_index)

                    mixdb.mixtures.append(MRecord(
                        name=None,
                        target_file_index=target_file_index,
                        target_augmentation_index=target_augmentation_index,
                        noise_file_index=noise_file_index,
                        noise_augmentation_index=noise_augmentation_index,
                        noise_offset=choice(range(noise_length)),
                        samples=target_length,
                        snr=snr.value,
                        spectral_mask_index=spectral_mask_index,
                        spectral_mask_seed=randint(0, np.iinfo('i').max),
                        random_snr=snr.is_random))

                    used_noise_samples += target_length

    return len(used_noise_files), used_noise_samples


def _get_next_noise_indices(mixdb: MixtureDatabase,
                            noise_file_index: int,
                            noise_augmentation_index: int) -> Tuple[int, int, int]:
    noise_augmentation_index += 1
    if noise_augmentation_index == len(mixdb.noise_augmentations):
        noise_augmentation_index = 0
        noise_file_index += 1
        if noise_file_index == len(mixdb.noises):
            noise_file_index = 0

    return (noise_file_index,
            noise_augmentation_index,
            mixdb.augmented_noise_length(noise_file_index, noise_augmentation_index))


def _get_next_noise_offset(mixdb: MixtureDatabase,
                           target_length: int,
                           noise_file_index: int,
                           noise_augmentation_index: int,
                           noise_offset: int) -> Tuple[int, int, int]:
    from sonusai import SonusAIError

    if noise_offset + target_length >= mixdb.augmented_noise_length(noise_file_index, noise_augmentation_index):
        if noise_offset == 0:
            raise SonusAIError('Length of target audio exceeds length of noise audio')

        noise_offset = 0
        noise_augmentation_index += 1
        if noise_augmentation_index == len(mixdb.noise_augmentations):
            noise_augmentation_index = 0
            noise_file_index += 1
            if noise_file_index == len(mixdb.noises):
                noise_file_index = 0

    return noise_file_index, noise_augmentation_index, noise_offset


def _get_target_info(mixdb: MixtureDatabase,
                     augmented_target_indices: List[int],
                     augmented_targets: AugmentedTargets) -> Tuple[List[int], List[int], int]:
    from sonusai.mixture import estimate_augmented_length_from_length

    target_file_index = []
    target_augmentation_index = []
    target_length = 0
    for idx in augmented_target_indices:
        tfi = augmented_targets[idx].target_file_index
        tai = augmented_targets[idx].target_augmentation_index

        target_file_index.append(tfi)
        target_augmentation_index.append(tai)

        target_length = max(estimate_augmented_length_from_length(length=mixdb.targets[tfi].samples,
                                                                  augmentation=mixdb.target_augmentations[tai],
                                                                  length_common_denominator=mixdb.feature_step_samples),
                            target_length)
    return target_file_index, target_augmentation_index, target_length



from dataclasses import dataclass
from dataclasses import field
from functools import lru_cache
from typing import Any
from typing import Dict
from typing import List
from typing import NamedTuple
from typing import Tuple
from typing import Union

from sonusai.mixture.types import AudioF
from sonusai.mixture.types import AudioT
from sonusai.mixture.types import AudiosF
from sonusai.mixture.types import AudiosT
from sonusai.mixture.types import Augmentation
from sonusai.mixture.types import Augmentations
from sonusai.mixture.types import ClassCount
from sonusai.mixture.types import DataClassSonusAIMixin
from sonusai.mixture.types import Feature
from sonusai.mixture.types import GeneralizedIDs
from sonusai.mixture.types import ImpulseResponseData
from sonusai.mixture.types import ImpulseResponseFiles
from sonusai.mixture.types import Location
from sonusai.mixture.types import NoiseFiles
from sonusai.mixture.types import Optional
from sonusai.mixture.types import Segsnr
from sonusai.mixture.types import SpectralMasks
from sonusai.mixture.types import TargetFile
from sonusai.mixture.types import TargetFiles
from sonusai.mixture.types import Truth
from sonusai.mixture.types import TruthSettings
from sonusai.mixture.types import UniversalSNR


# NOTE: global object is required for run-time performance; using 'partial' is much slower.
@dataclass
class MPGlobal:
    noise_augmentations: Optional[Augmentations] = None
    ir_data: Optional[List[ImpulseResponseData]] = None


MP_GLOBAL = MPGlobal()


@dataclass
class MRecord(DataClassSonusAIMixin):
    name: Optional[Union[Location]] = None
    noise_augmentation_index: Optional[int] = None
    noise_file_index: Optional[int] = None
    noise_offset: Optional[int] = None
    noise_snr_gain: Optional[float] = None
    random_snr: Optional[bool] = None
    samples: Optional[int] = None
    snr: Optional[float] = None
    spectral_mask_index: Optional[int] = None
    spectral_mask_seed: Optional[int] = None
    target_augmentation_index: Optional[List[int]] = None
    target_file_index: Optional[List[int]] = None
    target_gain: Optional[List[float]] = None
    target_snr_gain: Optional[float] = None


MRecords = List[MRecord]


@dataclass
class MixtureDatabaseConfig(DataClassSonusAIMixin):
    asr_manifest: List[Location] = field(default_factory=list)
    class_balancing: Optional[bool] = False
    class_balancing_augmentation: Optional[Augmentation] = None
    class_labels: Optional[List[str]] = None
    class_weights_threshold: Optional[List[float]] = None
    feature: Optional[str] = None
    first_cba_index: Optional[int] = None
    ir_files: Optional[ImpulseResponseFiles] = None
    mixtures: Optional[MRecords] = None
    noise_augmentations: Optional[Augmentations] = None
    noise_mix_mode: Optional[str] = 'exhaustive'
    noises: Optional[NoiseFiles] = None
    num_classes: Optional[int] = None
    random_snrs: Optional[List[str]] = None
    seed: Optional[int] = 0
    snrs: Optional[List[float]] = None
    spectral_masks: Optional[SpectralMasks] = None
    target_augmentations: Optional[Augmentations] = None
    targets: Optional[TargetFiles] = None
    truth_mutex: Optional[bool] = None
    truth_reduction_function: Optional[str] = None
    truth_settings: Optional[TruthSettings] = None


@dataclass
class TransformConfig:
    N: int
    R: int
    bin_start: int
    bin_end: int
    ttype: str


class MixtureDatabase:
    def __init__(self,
                 config: Union[Location, MixtureDatabaseConfig],
                 location: Optional[Location] = None,
                 lazy_load: bool = True,
                 use_cache: bool = True):
        from pyaaware import FeatureGenerator

        self._location: Location

        if isinstance(config, MixtureDatabaseConfig):
            self._config: MixtureDatabaseConfig = config
            self._location = location
        else:
            self._config = _load_from_location(config)
            self._location = config

        self.fg = FeatureGenerator(feature_mode=self.feature,
                                   num_classes=self.num_classes,
                                   truth_mutex=self.truth_mutex)

        self.ft_config = TransformConfig(N=self.fg.ftransform_N,
                                         R=self.fg.ftransform_R,
                                         bin_start=self.fg.bin_start,
                                         bin_end=self.fg.bin_end,
                                         ttype=self.fg.ftransform_ttype)

        self.eft_config = TransformConfig(N=self.fg.eftransform_N,
                                          R=self.fg.eftransform_R,
                                          bin_start=self.fg.bin_start,
                                          bin_end=self.fg.bin_end,
                                          ttype=self.fg.eftransform_ttype)

        self.it_config = TransformConfig(N=self.fg.itransform_N,
                                         R=self.fg.itransform_R,
                                         bin_start=self.fg.bin_start,
                                         bin_end=self.fg.bin_end,
                                         ttype=self.fg.itransform_ttype)

        self._ir_data: List[ImpulseResponseData] = []
        self._asr_manifest_data: Dict[str, str] = {}
        self._use_cache = use_cache

        if not lazy_load:
            self.load_ir_data()
            self.load_asr_data()

    @property
    def json(self) -> str:
        """Convert MixtureDatabase to JSON

        :return: JSON representation of database
        """
        return self._config.to_json(indent=2)

    def save(self) -> None:
        """Save the MixtureDatabase as a JSON file
        """
        from os.path import join

        json_name = join(self._location, 'mixdb.json')
        with open(file=json_name, mode='w') as file:
            file.write(self.json)

    def forward_transform(self, audio: AudioT) -> AudioF:
        """Transform time domain data into frequency domain using the forward transform config from the feature

        A new transform is used for each call; i.e., state is not maintained between calls to forward_transform().

        :param audio: Time domain data [samples]
        :return: Frequency domain data [frames, bins]
        """
        from pyaaware import AawareForwardTransform

        from sonusai.mixture import calculate_transform_from_audio

        audio_f, _ = calculate_transform_from_audio(audio=audio,
                                                    transform=AawareForwardTransform(N=self.ft_config.N,
                                                                                     R=self.ft_config.R,
                                                                                     bin_start=self.ft_config.bin_start,
                                                                                     bin_end=self.ft_config.bin_end,
                                                                                     ttype=self.ft_config.ttype))
        return audio_f

    def inverse_transform(self, transform: AudioF, trim: bool = True) -> AudioT:
        """Transform frequency domain data into time domain using the inverse transform config from the feature

        A new transform is used for each call; i.e., state is not maintained between calls to inverse_transform().

        :param transform: Frequency domain data [frames, bins]
        :param trim: Removes starting samples so output waveform will be time-aligned with input waveform to the
                     transform
        :return: Time domain data [samples]
        """
        import numpy as np

        from pyaaware import AawareInverseTransform

        from sonusai.mixture import calculate_audio_from_transform

        audio, _ = calculate_audio_from_transform(data=transform,
                                                  transform=AawareInverseTransform(N=self.it_config.N,
                                                                                   R=self.it_config.R,
                                                                                   bin_start=self.it_config.bin_start,
                                                                                   bin_end=self.it_config.bin_end,
                                                                                   ttype=self.it_config.ttype,
                                                                                   gain=np.float32(1)),
                                                  trim=trim)
        return audio

    @property
    def config(self) -> MixtureDatabaseConfig:
        return self._config

    @property
    def location(self) -> Location:
        return self._location

    @property
    def use_cache(self) -> bool:
        return self._use_cache

    @property
    def asr_manifest(self) -> List[Location]:
        return self._config.asr_manifest

    @asr_manifest.setter
    def asr_manifest(self, value: List[Location]) -> None:
        self._config.asr_manifest = value

    @property
    def class_balancing(self) -> bool:
        return self._config.class_balancing

    @class_balancing.setter
    def class_balancing(self, value: bool) -> None:
        self._config.class_balancing = value

    @property
    def class_balancing_augmentation(self) -> Augmentation:
        return self._config.class_balancing_augmentation

    @class_balancing_augmentation.setter
    def class_balancing_augmentation(self, value: Augmentation) -> None:
        self._config.class_balancing_augmentation = value

    @property
    def class_labels(self) -> List[str]:
        return self._config.class_labels

    @class_labels.setter
    def class_labels(self, value: List[str]) -> None:
        from sonusai import SonusAIError

        if value is not None and (not isinstance(value, list) or len(value) != self.num_classes):
            raise SonusAIError(f'invalid class_labels length')
        self._config.class_labels = value

    @property
    def class_weights_threshold(self) -> List[float]:
        return self._config.class_weights_threshold

    @class_weights_threshold.setter
    def class_weights_threshold(self, value: List[float]) -> None:
        from sonusai import SonusAIError

        if len(value) not in [1, self.num_classes]:
            raise SonusAIError(f'invalid class_weights_threshold length: {len(value)}')
        self._config.class_weights_threshold = value

    @property
    def noise_mix_mode(self) -> str:
        return self._config.noise_mix_mode

    @noise_mix_mode.setter
    def noise_mix_mode(self, value: str) -> None:
        self._config.noise_mix_mode = value

    @property
    def feature(self) -> str:
        return self._config.feature

    @feature.setter
    def feature(self, value: str) -> None:
        self._config.feature = value

    @property
    def transform_frame_ms(self) -> float:
        from sonusai.mixture import SAMPLE_RATE

        return float(self.ft_config.R) / float(SAMPLE_RATE / 1000)

    @property
    def feature_ms(self) -> float:
        return self.transform_frame_ms * self.fg.decimation * self.fg.stride

    @property
    def feature_samples(self) -> int:
        return self.ft_config.R * self.fg.decimation * self.fg.stride

    @property
    def feature_step_ms(self) -> float:
        return self.transform_frame_ms * self.fg.decimation * self.fg.step

    @property
    def feature_step_samples(self) -> int:
        return self.ft_config.R * self.fg.decimation * self.fg.step

    @property
    def first_cba_index(self) -> int:
        return self._config.first_cba_index

    @first_cba_index.setter
    def first_cba_index(self, value: int) -> None:
        self._config.first_cba_index = value

    @property
    def ir_files(self) -> ImpulseResponseFiles:
        return self._config.ir_files

    @property
    def mixtures(self) -> MRecords:
        return self._config.mixtures

    @mixtures.setter
    def mixtures(self, value: MRecords) -> None:
        self._config.mixtures = value

    @property
    def noise_augmentations(self) -> Augmentations:
        return self._config.noise_augmentations

    @noise_augmentations.setter
    def noise_augmentations(self, value: Augmentations) -> None:
        self._config.noise_augmentations = value

    @property
    def noises(self) -> NoiseFiles:
        return self._config.noises

    @noises.setter
    def noises(self, value: NoiseFiles) -> None:
        self._config.noises = value

    @property
    def num_classes(self) -> int:
        return self._config.num_classes

    @num_classes.setter
    def num_classes(self, value: int) -> None:
        self._config.num_classes = value

    @property
    def random_snrs(self) -> List[str]:
        return self._config.random_snrs

    @random_snrs.setter
    def random_snrs(self, value: List[str]) -> None:
        self._config.random_snrs = value

    @property
    def seed(self) -> int:
        return self._config.seed

    @seed.setter
    def seed(self, value: int) -> None:
        self._config.seed = value

    @property
    def snrs(self) -> List[float]:
        return self._config.snrs

    @snrs.setter
    def snrs(self, value: List[float]) -> None:
        self._config.snrs = value

    @property
    def all_snrs(self) -> List[UniversalSNR]:
        all_snrs: List[UniversalSNR] = []
        for snr in self.snrs:
            all_snrs.append(UniversalSNR(is_random=False, raw_value=snr))
        for random_snr in self.random_snrs:
            all_snrs.append(UniversalSNR(is_random=True, raw_value=random_snr))
        return all_snrs

    @property
    def spectral_masks(self) -> SpectralMasks:
        return self._config.spectral_masks

    @spectral_masks.setter
    def spectral_masks(self, value: SpectralMasks) -> None:
        self._config.spectral_masks = value

    @property
    def target_augmentations(self) -> Augmentations:
        return self._config.target_augmentations

    @target_augmentations.setter
    def target_augmentations(self, value: Augmentations) -> None:
        self._config.target_augmentations = value

    @property
    def targets(self) -> TargetFiles:
        return self._config.targets

    @targets.setter
    def targets(self, value: TargetFiles) -> None:
        self._config.targets = value

    @property
    def truth_mutex(self) -> bool:
        return self._config.truth_mutex

    @truth_mutex.setter
    def truth_mutex(self, value: bool) -> None:
        self._config.truth_mutex = value

    @property
    def truth_reduction_function(self) -> str:
        return self._config.truth_reduction_function

    @truth_reduction_function.setter
    def truth_reduction_function(self, value: str) -> None:
        self._config.truth_reduction_function = value

    @property
    def truth_settings(self) -> TruthSettings:
        return self._config.truth_settings

    @truth_settings.setter
    def truth_settings(self, value: TruthSettings) -> None:
        self._config.truth_settings = value

    def augmented_target_length(self, file_index: int,
                                augmentation_index: int) -> int:
        from sonusai.mixture import estimate_augmented_length_from_length

        return estimate_augmented_length_from_length(length=self.targets[file_index].samples,
                                                     augmentation=self.target_augmentations[augmentation_index],
                                                     length_common_denominator=self.feature_step_samples)

    def augmented_noise_length(self, file_index: int,
                               augmentation_index: int) -> int:
        from sonusai.mixture import estimate_augmented_length_from_length

        return estimate_augmented_length_from_length(length=self.noises[file_index].samples,
                                                     augmentation=self.noise_augmentations[augmentation_index])

    def raw_target_audio(self, file_index: int) -> AudioT:
        if self._use_cache:
            return self._cache_raw_target_audio(file_index)
        else:
            return self._no_cache_raw_target_audio(file_index)

    @lru_cache
    def _cache_raw_target_audio(self, file_index: int) -> AudioT:
        """Get raw target audio (using cache)

        :param file_index: Target audio file name
        :return: Raw target audio
        """
        return self._no_cache_raw_target_audio(file_index)

    def _no_cache_raw_target_audio(self, file_index: int) -> AudioT:
        """Get raw target audio (without using cache)

        :param file_index: Target audio file index
        :return: Raw target audio
        """
        from sonusai.mixture import read_audio

        return read_audio(self.targets[file_index].name)

    @property
    def raw_target_audio_cache_info(self) -> NamedTuple:
        return self._cache_raw_target_audio.cache_info()

    def augmented_target_audio(self, file_index: int, augmentation_index: int) -> AudioT:
        """Get augmented target audio

        :param file_index: Target audio file index
        :param augmentation_index: Target augmentation index
        :return: Augmented target audio
        """
        if self._use_cache:
            return self._cache_augmented_target_audio(file_index, augmentation_index)
        else:
            return self._no_cache_augmented_target_audio(file_index, augmentation_index)

    @lru_cache
    def _cache_augmented_target_audio(self, file_index: int, augmentation_index: int) -> AudioT:
        """Get augmented target audio (using cache)

        :param file_index: Target audio file name
        :param augmentation_index: Target augmentation index
        :return: Augmented target audio
        """
        return self._no_cache_augmented_target_audio(file_index, augmentation_index)

    def _no_cache_augmented_target_audio(self, file_index: int, augmentation_index: int) -> AudioT:
        """Get augmented target audio (without using cache)

        :param file_index: Target audio file index
        :param augmentation_index: Target augmentation index
        :return: Augmented target audio
        """
        from sonusai.mixture import apply_augmentation
        from sonusai.mixture import apply_ir
        from sonusai.mixture import read_audio

        audio = read_audio(self.targets[file_index].name)
        augmentation = self.target_augmentations[augmentation_index]
        audio = apply_augmentation(audio, augmentation, self.feature_step_samples)
        if augmentation.ir is not None:
            audio = apply_ir(audio, self.ir_data[int(augmentation.ir)])

        return audio

    @property
    def augmented_target_audio_cache_info(self) -> NamedTuple:
        return self._cache_augmented_target_audio.cache_info()

    def augmented_noise_audio(self, file_index: int, augmentation_index: int) -> AudioT:
        """Get augmented noise audio

        :param file_index: Noise audio file index
        :param augmentation_index: Noise augmentation index
        :return: Augmented noise audio
        """
        if self._use_cache:
            return self._cache_augmented_noise_audio(file_index, augmentation_index)
        else:
            return self._no_cache_augmented_noise_audio(file_index, augmentation_index)

    @lru_cache
    def _cache_augmented_noise_audio(self, file_index: int, augmentation_index: int) -> AudioT:
        """Get augmented noise audio (using cache)

        :param file_index: Noise audio file index
        :param augmentation_index: Noise augmentation index
        :return: Augmented noise audio
        """
        return self._no_cache_augmented_noise_audio(file_index, augmentation_index)

    def _no_cache_augmented_noise_audio(self, file_index: int, augmentation_index: int) -> AudioT:
        """Get augmented noise audio (without using cache)

        :param file_index: Noise audio file index
        :param augmentation_index: Noise augmentation index
        :return: Augmented noise audio
        """
        from sonusai.mixture import apply_augmentation
        from sonusai.mixture import apply_ir
        from sonusai.mixture import read_audio

        audio = read_audio(self.noises[file_index].name)
        augmentation = self.noise_augmentations[augmentation_index]
        audio = apply_augmentation(audio, augmentation)
        if augmentation.ir is not None:
            audio = apply_ir(audio, self.ir_data[int(augmentation.ir)])

        return audio

    @property
    def augmented_noise_audio_cache_info(self) -> NamedTuple:
        return self._cache_augmented_noise_audio.cache_info()

    @property
    def ir_data(self) -> List[ImpulseResponseData]:
        """Get the list of impulse response data, loading from disk if necessary

        :return: List of impulse response data
        """
        if not self._ir_data:
            self.load_ir_data()

        return self._ir_data

    def load_ir_data(self) -> None:
        """Load all impulse response data into memory

        :return: List of impulse response data
        """
        from os.path import join

        from sonusai.mixture import generate_coefficients_filename
        from sonusai.mixture import read_ir
        from sonusai.mixture import save_ir_data
        from sonusai.utils import max_text_width

        if len(self.ir_files) == 0:
            self._ir_data = None
            return

        self._ir_data = []
        width = max_text_width(len(self.ir_files))
        for idx, ir_file in enumerate(self.ir_files):
            raw = read_ir(ir_file)
            name = join(self._location, generate_coefficients_filename(idx, width))
            save_ir_data(name, raw.filter)
            self._ir_data.append(ImpulseResponseData(
                name=ir_file,
                sample_rate=raw.sample_rate,
                offset=raw.offset,
                length=len(raw.filter),
                coefficients_file=name))

    def target_asr_data(self, target: TargetFile) -> Union[str, None]:
        """Get the ASR data for the given target, loading from disk if necessary

        :param target: TargetFile
        :return: ASR text or None
        """
        from sonusai.mixture import tokenized_expand

        if not self._asr_manifest_data:
            self.load_asr_data()

        name, _ = tokenized_expand(target.name)
        return self._asr_manifest_data.get(name, None)

    def mixid_asr_data(self, mixid: int) -> List[Union[str, None]]:
        """Get the ASR data for the given mixid, loading from disk if necessary

        :param mixid: Mixture ID
        :return: List of ASR text or None
        """
        if not self._asr_manifest_data:
            self.load_asr_data()

        return [self.target_asr_data(self.targets[tfi]) for tfi in self.mixtures[mixid].target_file_index]

    def load_asr_data(self) -> None:
        """Load ASR data into memory

        Each line of a manifest file should be in the following format:

        {"audio_filepath": "/path/to/audio.wav", "text": "the transcription of the utterance", "duration": 23.147}

        The audio_filepath field should provide an absolute path to the audio file corresponding to the utterance. The
        text field should contain the full transcript for the utterance, and the duration field should reflect the
        duration of the utterance in seconds.

        Each entry in the manifest (describing one audio file) should be bordered by '{' and '}' and must be contained
        on one line. The fields that describe the file should be separated by commas, and have the form
        "field_name": value, as shown above.

        Since the manifest specifies the path for each utterance, the audio files do not have to be located in the same
        directory as the manifest, or even in any specific directory structure.

        The manifest dictionary consists of key/value pairs where the keys are target file names and the values are ASR
        text.
        """
        import json

        from sonusai import SonusAIError
        from sonusai.mixture import tokenized_expand
        from sonusai.mixture import Location

        expected_keys = ['audio_filepath', 'text', 'duration']

        def _error_preamble(e_name: Location, e_line_num: int) -> str:
            return f'Invalid entry in ASR manifest {e_name} line {e_line_num}'

        self._asr_manifest_data = {}

        for name in self._config.asr_manifest:
            with open(file=name, mode='r') as f:
                line_num = 1
                for line in f:
                    result = json.loads(line.strip())

                    for key in expected_keys:
                        if key not in result:
                            SonusAIError(f'{_error_preamble(name, line_num)}: missing field "{key}"')

                    for key in result.keys():
                        if key not in expected_keys:
                            SonusAIError(f'{_error_preamble(name, line_num)}: unknown field "{key}"')

                    key, _ = tokenized_expand(result['audio_filepath'])
                    value = result['text']

                    if key in self._asr_manifest_data:
                        SonusAIError(f'{_error_preamble(name, line_num)}: entry already exists')

                    self._asr_manifest_data[key] = value

                    line_num += 1

    @property
    def augmented_target_samples(self) -> int:
        from itertools import product

        it = list(product(*[range(len(self.targets)), range(len(self.target_augmentations))]))
        return sum([self.augmented_target_length(fi, ai) for fi, ai in it])

    @property
    def augmented_noise_samples(self) -> int:
        from itertools import product

        it = list(product(*[range(len(self.noises)), range(len(self.noise_augmentations))]))
        return sum([self.augmented_noise_length(fi, ai) for fi, ai in it])

    def total_samples(self, mixids: GeneralizedIDs = '*') -> int:
        return sum([self.mixture_samples(mixid) for mixid in self.mixids_to_list(mixids)])

    def total_transform_frames(self, mixids: GeneralizedIDs = '*') -> int:
        return self.total_samples(mixids) // self.ft_config.R

    def total_feature_frames(self, mixids: GeneralizedIDs = '*') -> int:
        return self.total_samples(mixids) // self.feature_step_samples

    def mixture_samples(self, mixid: int) -> int:
        return self.mixtures[mixid].samples

    def mixture_transform_frames(self, mixid: int) -> int:
        return self.mixture_samples(mixid) // self.ft_config.R

    def mixture_feature_frames(self, mixid: int) -> int:
        return self.mixture_samples(mixid) // self.feature_step_samples

    def mixids_to_list(self, ids: Optional[GeneralizedIDs] = None) -> List[int]:
        """Resolve generalized mixture IDs to a list of integers

        :param ids: Generalized mixture IDs
        :return: List of mixture ID integers
        """
        return generic_ids_to_list(len(self.mixtures), ids)

    def mixture_metadata(self, mixid: int) -> str:
        """Create a string of metadata for a mixture ID

        :param mixid: Mixture ID
        :return: String of metadata
        """
        mrecord = self.mixtures[mixid]
        metadata = ''
        for ti in range(len(mrecord.target_file_index)):
            tfi = mrecord.target_file_index[ti]
            tai = mrecord.target_augmentation_index[ti]
            metadata += f'target {ti} name: {self.targets[tfi].name}\n'
            metadata += f'target {ti} augmentation: {self.target_augmentations[tai].to_dict()}\n'
            if self.target_augmentations[tai].ir is None:
                ir_name = None
            else:
                ir_name = self.ir_files[int(self.target_augmentations[tai].ir)]
            metadata += f'target {ti} ir: {ir_name}\n'
            metadata += f'target {ti} target_gain: {mrecord.target_gain[ti]}\n'
            truth_settings = self.targets[tfi].truth_settings
            for tsi in range(len(truth_settings)):
                metadata += f'target {ti} truth index {tsi}: {truth_settings[tsi].index}\n'
                metadata += f'target {ti} truth function {tsi}: {truth_settings[tsi].function}\n'
                metadata += f'target {ti} truth config {tsi}: {truth_settings[tsi].config}\n'
            metadata += f'target {ti} asr: {self.target_asr_data(self.targets[tfi])}\n'
        nfi = mrecord.noise_file_index
        nai = mrecord.noise_augmentation_index
        metadata += f'noise name: {self.noises[nfi].name}\n'
        metadata += f'noise augmentation: {self.noise_augmentations[nai].to_dict()}\n'
        if self.noise_augmentations[nai].ir is None:
            ir_name = None
        else:
            ir_name = self.ir_files[int(self.noise_augmentations[nai].ir)]
        metadata += f'noise ir: {ir_name}\n'
        metadata += f'noise offset: {mrecord.noise_offset}\n'
        metadata += f'snr: {mrecord.snr}\n'
        metadata += f'random_snr: {mrecord.random_snr}\n'
        metadata += f'samples: {mrecord.samples}\n'
        metadata += f'target_snr_gain: {mrecord.target_snr_gain}\n'
        metadata += f'noise_snr_gain: {mrecord.noise_snr_gain}\n'

        return metadata

    def write_mixture_metadata(self, mixid: int) -> None:
        """Write mixture metadata to a text file

        :param mixid: Mixture ID
        """
        from os.path import splitext

        with open(file=splitext(self.mixture_filename(mixid))[0] + '.txt', mode='w') as f:
            f.write(self.mixture_metadata(mixid))

    def mixture_filename(self, mixid: int) -> Location:
        """Get the HDF5 file name for the given mixture ID

        :param mixid: Mixture ID
        :return: File name
        """
        from os.path import join

        return join(self._location, self.mixtures[mixid].name) if self._location is not None else None

    def check_audio_files_exist(self) -> None:
        """Walk through all the noise and target audio files in a mixture database ensuring that they exist
        """
        from os.path import exists

        from sonusai import SonusAIError
        from sonusai.mixture import tokenized_expand

        for file_index in range(len(self.noises)):
            file_name, _ = tokenized_expand(self.noises[file_index].name)
            if not exists(file_name):
                raise SonusAIError(f'Could not find {file_name}')

        for file_index in range(len(self.targets)):
            file_name, _ = tokenized_expand(self.targets[file_index].name)
            if not exists(file_name):
                raise SonusAIError(f'Could not find {file_name}')

    def read_mixture_data(self, mixid: int, items: Union[List[str], str]) -> Any:
        """Read mixture data from a mixture HDF5 file

        :param mixid: Mixture ID
        :param items: String(s) of dataset(s) to retrieve
        :return: Data (or tuple of data)
        """
        from os.path import exists
        from typing import Any

        import h5py
        import numpy as np

        from sonusai import SonusAIError

        def _get_dataset(file: h5py.File, d_name: str) -> Any:
            if d_name in file:
                return np.array(file[d_name])
            return None

        if not isinstance(items, list):
            items = [items]

        name = self.mixture_filename(mixid)
        if exists(name):
            try:
                with h5py.File(name, 'r') as f:
                    result = ([_get_dataset(f, item) for item in items])
            except Exception as e:
                raise SonusAIError(f'Error reading {name}: {e}')
        else:
            result = ([None for _ in items])

        if len(items) == 1:
            result = result[0]

        return result

    def write_mixture_data(self, mixid: int, items: Union[List[Tuple[str, Any]], Tuple[str, Any]]) -> None:
        """Write mixture data to a mixture HDF5 file

        :param mixid: Mixture ID
        :param items: Tuple(s) of (name, data)
        """
        import h5py

        if not isinstance(items, list):
            items = [items]

        with h5py.File(self.mixture_filename(mixid), 'a') as f:
            for item in items:
                if item[0] in f:
                    del f[item[0]]
                f.create_dataset(name=item[0], data=item[1])

    def mixture_targets(self,
                        mixid: int,
                        force: bool = False) -> AudiosT:
        """Get the list of augmented target audio data (one per target in the mixup) for the given mixid

        :param mixid: Mixture ID
        :param force: Force computing data from original sources regardless of whether or not cached data exists
        :return: List of augmented target audio data (one per target in the mixup)
        """
        from sonusai.mixture import apply_augmentation
        from sonusai.mixture import apply_gain
        from sonusai.mixture import pad_audio_to_length

        if not force:
            targets = self.read_mixture_data(mixid, 'targets')
            if targets is not None:
                return list(targets)

        mrecord = self.mixtures[mixid]
        targets = []
        for idx in range(len(mrecord.target_file_index)):
            target = self.raw_target_audio(mrecord.target_file_index[idx])
            target = apply_augmentation(audio=target,
                                        augmentation=self.target_augmentations[mrecord.target_augmentation_index[idx]],
                                        length_common_denominator=self.feature_step_samples)
            target = apply_gain(audio=target, gain=mrecord.target_snr_gain)
            target = pad_audio_to_length(audio=target, length=mrecord.samples)
            targets.append(target)

        return targets

    def mixture_targets_f(self,
                          mixid: int,
                          targets: Optional[AudiosT] = None,
                          force: bool = False) -> AudiosF:
        """Get the list of augmented target transform data (one per target in the mixup) for the given mixid

        :param mixid: Mixture ID
        :param targets: List of augmented target audio data (one per target in the mixup) for the given mixid
        :param force: Force computing data from original sources regardless of whether or not cached data exists
        :return: List of augmented target transform data (one per target in the mixup)
        """
        if targets is None:
            targets = self.mixture_targets(mixid=mixid, force=force)

        return [self.forward_transform(target) for target in targets]

    def mixture_target(self,
                       mixid: int,
                       targets: Optional[AudiosT] = None,
                       force: bool = False) -> AudioT:
        """Get the augmented target audio data for the given mixid

        :param mixid: Mixture ID
        :param targets: List of augmented target audio data (one per target in the mixup) for the given mixid
        :param force: Force computing data from original sources regardless of whether or not cached data exists
        :return: Augmented target audio data
        """
        import numpy as np

        from sonusai.mixture import apply_ir

        if not force:
            target = self.read_mixture_data(mixid, 'target')
            if target is not None:
                return target

        if targets is None:
            targets = self.mixture_targets(mixid=mixid, force=force)

        # Apply impulse responses to targets
        targets_ir = []
        for idx, target in enumerate(targets):
            ir_idx = self.target_augmentations[self.mixtures[mixid].target_augmentation_index[idx]].ir
            if ir_idx is not None:
                targets_ir.append(apply_ir(audio=target, ir=self.ir_data[int(ir_idx)]))
            else:
                targets_ir.append(target)
        targets = targets_ir

        return np.sum(targets, axis=0)

    def mixture_target_f(self,
                         mixid: int,
                         targets: Optional[AudiosT] = None,
                         target: Optional[AudioT] = None,
                         force: bool = False) -> AudioF:
        """Get the augmented target transform data for the given mixid

        :param mixid: Mixture ID
        :param targets: List of augmented target audio data (one per target in the mixup) for the given mixid
        :param target: Augmented target audio for the given mixid
        :param force: Force computing data from original sources regardless of whether or not cached data exists
        :return: Augmented target transform data
        """
        if target is None:
            target = self.mixture_target(mixid=mixid, targets=targets, force=force)

        return self.forward_transform(target)

    def mixture_noise(self,
                      mixid: int,
                      force: bool = False) -> AudioT:
        """Get the augmented noise audio data for the given mixid

        :param mixid: Mixture ID
        :param force: Force computing data from original sources regardless of whether or not cached data exists
        :return: Augmented noise audio data
        """
        from sonusai.mixture import apply_gain
        from sonusai.mixture import get_next_noise

        if not force:
            noise = self.read_mixture_data(mixid, 'noise')
            if noise is not None:
                return noise

        mrecord = self.mixtures[mixid]
        noise = self.augmented_noise_audio(mrecord.noise_file_index, mrecord.noise_augmentation_index)
        noise = get_next_noise(audio=noise, offset=mrecord.noise_offset, length=mrecord.samples)
        noise = apply_gain(audio=noise, gain=mrecord.noise_snr_gain)

        return noise

    def mixture_noise_f(self,
                        mixid: int,
                        noise: Optional[AudioT] = None,
                        force: bool = False) -> AudioF:
        """Get the augmented noise transform for the given mixid

        :param mixid: Mixture ID
        :param noise: Augmented noise audio data for the given mixid
        :param force: Force computing data from original sources regardless of whether or not cached data exists
        :return: Augmented noise transform data
        """
        if noise is None:
            noise = self.mixture_noise(mixid=mixid, force=force)

        return self.forward_transform(noise)

    def mixture_mixture(self,
                        mixid: int,
                        targets: Optional[AudiosT] = None,
                        target: Optional[AudioT] = None,
                        noise: Optional[AudioT] = None,
                        force: bool = False) -> AudioT:
        """Get the mixture audio data for the given mixid

        :param mixid: Mixture ID
        :param targets: List of augmented target audio data (one per target in the mixup) for the given mixid
        :param target: Augmented target audio data for the given mixid
        :param noise: Augmented noise audio data for the given mixid
        :param force: Force computing data from original sources regardless of whether cached data exists
        :return: Mixture audio data
        """
        if not force:
            mixture = self.read_mixture_data(mixid, 'mixture')
            if mixture is not None:
                return mixture

        if target is None:
            target = self.mixture_target(mixid=mixid, targets=targets)

        if noise is None:
            noise = self.mixture_noise(mixid=mixid)

        mixture = target + noise

        return mixture

    def mixture_mixture_f(self, mixid: int,
                          targets: Optional[AudiosT] = None,
                          target: Optional[AudioT] = None,
                          noise: Optional[AudioT] = None,
                          mixture: Optional[AudioT] = None,
                          force: bool = False) -> AudioF:
        """Get the mixture transform for the given mixid

        :param mixid: Mixture ID
        :param targets: List of augmented target audio data (one per target in the mixup) for the given mixid
        :param target: Augmented target audio data for the given mixid
        :param noise: Augmented noise audio data for the given mixid
        :param mixture: Mixture audio data for the given mixid
        :param force: Force computing data from original sources regardless of whether cached data exists
        :return: Mixture transform data
        """
        if mixture is None:
            mixture = self.mixture_mixture(mixid=mixid, targets=targets, target=target, noise=noise, force=force)

        return self.forward_transform(mixture)

    def mixture_truth_t(self,
                        mixid: int,
                        targets: Optional[AudiosT] = None,
                        noise: Optional[AudioT] = None,
                        force: bool = False) -> Truth:
        """Get the truth_t data for the given mixid

        :param mixid: Mixture ID
        :param targets: List of augmented target audio data (one per target in the mixup) for the given mixid
        :param noise: Augmented noise audio data for the given mixid
        :param force: Force computing data from original sources regardless of whether cached data exists
        :return: truth_t data
        """
        import numpy as np

        from sonusai import SonusAIError
        from sonusai.mixture import TruthFunctionConfig
        from sonusai.mixture import truth_function

        if not force:
            truth_t = self.read_mixture_data(mixid, 'truth_t')
            if truth_t is not None:
                return truth_t

        if targets is None:
            targets = self.mixture_targets(mixid=mixid)

        if noise is None:
            noise = self.mixture_noise(mixid=mixid)

        mrecord = self.mixtures[mixid]
        if len(targets) != len(mrecord.target_file_index):
            raise SonusAIError('Number of target audio entries does not match number of targets')

        if not all(len(target_audio) == len(noise) for target_audio in targets):
            raise SonusAIError('Lengths of target audio do not match length of noise audio')

        truth_t = np.zeros((self.mixture_samples(mixid), self.num_classes), dtype=np.float32)
        for idx in range(len(targets)):
            for truth_setting in self.targets[mrecord.target_file_index[idx]].truth_settings:
                config = TruthFunctionConfig(
                    feature=self.feature,
                    index=truth_setting.index,
                    function=truth_setting.function,
                    config=truth_setting.config,
                    num_classes=self.num_classes,
                    mutex=self.truth_mutex,
                    target_gain=mrecord.target_gain[idx] * mrecord.target_snr_gain
                )
                truth_t += truth_function(target_audio=targets[idx], noise_audio=noise, config=config)

        return truth_t

    def mixture_segsnr_t(self,
                         mixid: int,
                         targets: Optional[AudiosT] = None,
                         target: Optional[AudioT] = None,
                         noise: Optional[AudioT] = None,
                         force: bool = False) -> Segsnr:
        """Get the segsnr_t data for the given mixid

        :param mixid: Mixture ID
        :param targets: List of augmented target audio data (one per target in the mixup) for the given mixid
        :param target: Augmented target audio data for the given mixid
        :param noise: Augmented noise audio data for the given mixid
        :param force: Force computing data from original sources regardless of whether or not cached data exists
        :return: segsnr_t data
        """
        import numpy as np
        from pyaaware import AawareForwardTransform

        from sonusai import SonusAIError

        if not force:
            segsnr_t = self.read_mixture_data(mixid, 'segsnr_t')
            if segsnr_t is not None:
                return segsnr_t

        if target is None:
            target = self.mixture_target(mixid=mixid, targets=targets)

        if noise is None:
            noise = self.mixture_noise(mixid=mixid)

        fft = AawareForwardTransform(N=self.ft_config.N,
                                     R=self.ft_config.R,
                                     bin_start=self.ft_config.bin_start,
                                     bin_end=self.ft_config.bin_end,
                                     ttype=self.ft_config.ttype)

        segsnr_t = np.empty(self.mixture_samples(mixid), dtype=np.float32)

        _, target_energy = fft.execute_all(target)
        _, noise_energy = fft.execute_all(noise)

        offsets = range(0, self.mixture_samples(mixid), self.ft_config.R)
        if len(target_energy) != len(offsets):
            raise SonusAIError(f'Number of frames in energy, {len(target_energy)},'
                               f' is not number of frames in mixture, {len(offsets)}')

        for idx, offset in enumerate(offsets):
            indices = slice(offset, offset + self.ft_config.R)

            if noise_energy[idx] == 0:
                snr = np.float32(np.inf)
            else:
                snr = np.float32(target_energy[idx] / noise_energy[idx])

            segsnr_t[indices] = snr

        return segsnr_t

    def mixture_segsnr(self,
                       mixid: int,
                       segsnr_t: Optional[Segsnr] = None,
                       targets: Optional[AudiosT] = None,
                       target: Optional[AudioT] = None,
                       noise: Optional[AudioT] = None,
                       force: bool = False) -> Segsnr:
        """Get the segsnr data for the given mixid

        :param mixid: Mixture ID
        :param segsnr_t: segsnr_t data for the given mixid
        :param targets: List of augmented target audio data (one per target in the mixup) for the given mixid
        :param target: Augmented target audio data for the given mixid
        :param noise: Augmented noise audio data for the given mixid
        :param force: Force computing data from original sources regardless of whether or not cached data exists
        :return: segsnr data
        """
        if not force:
            segsnr = self.read_mixture_data(mixid, 'segsnr')
            if segsnr is not None:
                return segsnr

            segsnr_t = self.read_mixture_data(mixid, 'segsnr_t')
            if segsnr_t is not None:
                return segsnr_t[0::self.ft_config.R]

        if segsnr_t is None:
            segsnr_t = self.mixture_segsnr_t(mixid=mixid, targets=targets, target=target, noise=noise)

        segsnr = segsnr_t[0::self.ft_config.R]

        return segsnr

    def mixture_ft(self,
                   mixid: int,
                   targets: Optional[AudiosT] = None,
                   target: Optional[AudioT] = None,
                   noise: Optional[AudioT] = None,
                   mixture_f: Optional[AudioF] = None,
                   mixture: Optional[AudioT] = None,
                   truth_t: Optional[Truth] = None,
                   force: bool = False) -> Tuple[Feature, Truth]:
        """Get the feature and truth_f data for the given mixid

        :param mixid: Mixture ID
        :param targets: List of augmented target audio data (one per target in the mixup) for the given mixid
        :param target: Augmented target audio data for the given mixid
        :param noise: Augmented noise audio data for the given mixid
        :param mixture_f: Mixture transform data for the given mixid
        :param mixture: Mixture audio data for the given mixid
        :param truth_t: truth_t for the given mixid
        :param force: Force computing data from original sources regardless of whether or not cached data exists
        :return: Tuple of (feature, truth_f) data
        """
        import numpy as np

        from sonusai.mixture import apply_spectral_mask
        from sonusai.mixture import truth_reduction

        if not force:
            feature, truth_f = self.read_mixture_data(mixid, ['feature', 'truth_f'])
            if feature is not None and truth_f is not None:
                return feature, truth_f

        if mixture_f is None:
            mixture_f = self.mixture_mixture_f(mixid=mixid,
                                               targets=targets,
                                               target=target,
                                               noise=noise,
                                               mixture=mixture)

        if truth_t is None:
            truth_t = self.mixture_truth_t(mixid=mixid, targets=targets, noise=noise)

        mrecord = self.mixtures[mixid]

        transform_frames = self.mixture_transform_frames(mixid)
        feature_frames = self.mixture_feature_frames(mixid)

        if truth_t is None:
            truth_t = np.zeros((self.mixture_samples(mixid), self.num_classes), dtype=np.float32)

        feature = np.empty((feature_frames, self.fg.stride, self.fg.num_bands), dtype=np.float32)
        truth_f = np.empty((feature_frames, self.num_classes), dtype=np.complex64)

        feature_frame = 0
        for transform_frame in range(transform_frames):
            indices = slice(transform_frame * self.ft_config.R,
                            (transform_frame + 1) * self.ft_config.R)
            self.fg.execute(mixture_f[transform_frame],
                            truth_reduction(truth_t[indices], self.truth_reduction_function))

            if self.fg.eof():
                feature[feature_frame] = self.fg.feature()
                truth_f[feature_frame] = self.fg.truth()
                feature_frame += 1

        if np.isreal(truth_f).all():
            truth_f = np.float32(np.real(truth_f))

        if mrecord.spectral_mask_index is not None:
            feature = apply_spectral_mask(feature=feature,
                                          spectral_mask=self.spectral_masks[mrecord.spectral_mask_index],
                                          seed=mrecord.spectral_mask_seed)

        return feature, truth_f

    def mixture_feature(self,
                        mixid: int,
                        targets: Optional[AudiosT] = None,
                        noise: Optional[AudioT] = None,
                        mixture: Optional[AudioT] = None,
                        truth_t: Optional[Truth] = None,
                        force: bool = False) -> Feature:
        """Get the feature data for the given mixid

        :param mixid: Mixture ID
        :param targets: List of augmented target audio data (one per target in the mixup) for the given mixid
        :param noise: Augmented noise audio data for the given mixid
        :param mixture: Mixture audio data for the given mixid
        :param truth_t: truth_t for the given mixid
        :param force: Force computing data from original sources regardless of whether or not cached data exists
        :return: Feature data
        """
        feature, _ = self.mixture_ft(mixid=mixid,
                                     targets=targets,
                                     noise=noise,
                                     mixture=mixture,
                                     truth_t=truth_t,
                                     force=force)
        return feature

    def mixture_truth_f(self,
                        mixid: int,
                        targets: Optional[AudiosT] = None,
                        noise: Optional[AudioT] = None,
                        mixture: Optional[AudioT] = None,
                        truth_t: Optional[Truth] = None,
                        force: bool = False) -> Truth:
        """Get the truth_f data for the given mixid

        :param mixid: Mixture ID
        :param targets: List of augmented target audio data (one per target in the mixup) for the given mixid
        :param noise: Augmented noise audio data for the given mixid
        :param mixture: Mixture audio data for the given mixid
        :param truth_t: truth_t for the given mixid
        :param force: Force computing data from original sources regardless of whether or not cached data exists
        :return: truth_f data
        """
        _, truth_f = self.mixture_ft(mixid=mixid,
                                     targets=targets,
                                     noise=noise,
                                     mixture=mixture,
                                     truth_t=truth_t,
                                     force=force)
        return truth_f

    def mixture_class_count(self,
                            mixid: int,
                            targets: Optional[AudiosT] = None,
                            noise: Optional[AudioT] = None,
                            truth_t: Optional[Truth] = None) -> ClassCount:
        """Compute the number of samples for which each truth index is active for the given mixid

        :param mixid: Mixture ID
        :param targets: List of augmented target audio (one per target in the mixup) for the given mixid
        :param noise: Augmented noise audio for the given mixid
        :param truth_t: truth_t for the given mixid
        :return: List of class counts
        """
        import numpy as np

        if truth_t is None:
            truth_t = self.mixture_truth_t(mixid=mixid, targets=targets, noise=noise)

        class_count = [0] * self.num_classes
        num_classes = self.num_classes
        if self.truth_mutex:
            num_classes -= 1
        for cl in range(num_classes):
            class_count[cl] = int(np.sum(truth_t[:, cl] >= self.class_weights_threshold[cl]))

        return class_count


def get_mrecords_from_mixids(mixdb: MixtureDatabase, mixids: Optional[GeneralizedIDs] = None) -> MRecords:
    """Get a list of MRecords for the given mixture IDs

    :param mixdb: Mixture database
    :param mixids: Mixture IDs
    :return: MRecords
    """
    from copy import deepcopy

    return [deepcopy(mixdb.mixtures[i]) for i in mixdb.mixids_to_list(mixids)]


def _augment_noise_audio(audio: AudioT) -> AudiosT:
    from sonusai.mixture import apply_augmentation
    from sonusai.mixture import apply_ir

    results = []
    for augmentation in MP_GLOBAL.noise_augmentations:
        result = apply_augmentation(audio, augmentation)
        if augmentation.ir is not None:
            result = apply_ir(result, MP_GLOBAL.ir_data[int(augmentation.ir)])
        results.append(result)

    return results


def generic_ids_to_list(num_ids: int, ids: GeneralizedIDs = None) -> List[int]:
    """Resolve generalized IDs to a list of integers

    :param num_ids: Total number of indices
    :param ids: Generalized  IDs
    :return: List of ID integers
    """
    from sonusai import SonusAIError

    all_ids = list(range(num_ids))

    if ids is None:
        return all_ids

    if isinstance(ids, str):
        if ids == '*':
            return all_ids

        try:
            result = eval(f'{all_ids}[{ids}]')
            if not isinstance(result, list):
                result = [result]
            return result
        except NameError:
            raise SonusAIError(f'Empty ids {ids}')

    if isinstance(ids, range):
        result = list(ids)
    elif isinstance(ids, int):
        result = [ids]
    else:
        result = ids

    if not all(isinstance(x, int) and 0 <= x < num_ids for x in result):
        raise SonusAIError(f'Invalid entries in ids of {ids}')

    if not result:
        raise SonusAIError(f'Empty ids {ids}')

    return result


def _load_from_location(location: Location) -> MixtureDatabaseConfig:
    import json
    from os.path import exists
    from os.path import isdir
    from os.path import join

    from sonusai import SonusAIError

    if not isdir(location):
        raise SonusAIError(f'{location} is not a directory')

    filename = join(location, 'mixdb.json')
    if not exists(filename):
        raise SonusAIError(f'could not find mixture database in {location}')

    with open(file=filename, mode='r', encoding='utf-8') as f:
        return MixtureDatabaseConfig.from_dict(json.loads(f.read()))



from sonusai.utils.asr_manifest_functions.data import PathInfo
from sonusai.utils.asr_manifest_functions.data import TranscriptData
from sonusai.utils.asr_manifest_functions.librispeech import collect_librispeech_transcripts
from sonusai.utils.asr_manifest_functions.librispeech import get_librispeech_manifest_entry
from sonusai.utils.asr_manifest_functions.vctk_noisy_speech import collect_vctk_noisy_speech_transcripts
from sonusai.utils.asr_manifest_functions.vctk_noisy_speech import get_vctk_noisy_speech_manifest_entry
